__author__ = 'Lene Preuss <lene.preuss@gmail.com>'

import logging
from argparse import Namespace
from collections import defaultdict
from dataclasses import dataclass
from itertools import combinations
from multiprocessing.pool import ThreadPool as Pool
from pathlib import Path
from typing import Dict, List, Iterator, Optional, Tuple

from imagehash import ImageHash
from PIL import Image

from duplicate_images.common import path_with_parent
from duplicate_images.function_types import Cache, CacheEntry, HashFunction, ImagePair, Results
from duplicate_images.methods import get_hash_size_kwargs
from duplicate_images.progress_bar_manager import ProgressBarManager


@dataclass(frozen=True)
class PairFinderOptions:
    max_distance: int = 0
    hash_size: Optional[int] = None
    show_progress_bars: bool = False
    parallel: bool = False
    serial: bool = False

    @classmethod
    def from_args(cls, args: Namespace):
        return cls(args.max_distance, args.hash_size, args.progress, args.parallel, args.serial)


class ImagePairFinder:

    @classmethod
    def create(
            cls, files: List[Path], hash_algorithm: HashFunction,
            options: PairFinderOptions = PairFinderOptions(),
            hash_store: Optional[Cache] = None
    ) -> 'ImagePairFinder':
        if options.parallel:
            return ParallelImagePairFinder(
                files, hash_algorithm, options=options, hash_store=hash_store
            )
        if options.max_distance == 0 and not options.serial:
            return DictImagePairFinder(
                files, hash_algorithm, options=options, hash_store=hash_store
            )
        return SerialImagePairFinder(files, hash_algorithm, options=options, hash_store=hash_store)

    def __init__(  # pylint: disable = too-many-arguments
            self, files: List[Path], hash_algorithm: HashFunction,
            options: PairFinderOptions = PairFinderOptions(),
            hash_store: Optional[Cache] = None
    ) -> None:
        logging.info('Using %s', self.__class__.__name__)
        self.files = files
        self.algorithm = hash_algorithm
        self.hash_size_kwargs = get_hash_size_kwargs(hash_algorithm, options.hash_size)
        self.max_distance = options.max_distance
        self.hash_store = hash_store if hash_store is not None else {}
        self.progress_bars = ProgressBarManager.create(len(files), options.show_progress_bars)

    def get_pairs(self) -> Results:
        raise NotImplementedError()

    def get_hash(self, file: Path) -> CacheEntry:
        self.progress_bars.update_reader()
        try:
            cached = self.hash_store.get(file)
            if cached is not None:
                return file, cached

            image_hash = self.algorithm(Image.open(file), **self.hash_size_kwargs)
            self.hash_store[file] = image_hash
            return file, image_hash
        except OSError as err:
            logging.warning('%s: %s', path_with_parent(file), err)
            return file, None

    def precalculate_hashes(self, image_files: List[Path]) -> List[CacheEntry]:
        return [self.get_hash(file) for file in image_files]


class SerialImagePairFinder(ImagePairFinder):

    def __init__(  # pylint: disable = too-many-arguments
            self, files: List[Path], hash_algorithm: HashFunction,
            options: PairFinderOptions = PairFinderOptions(),
            hash_store: Optional[Cache] = None
    ) -> None:
        super().__init__(files, hash_algorithm, options, hash_store)
        self.precalculated_hashes = self.get_hashes(files)
        self.progress_bars.close_reader()

    def get_pairs(self) -> Results:
        image_files = list(self.precalculated_hashes.keys())
        logging.info('%d hashes calculated', len(self.precalculated_hashes))
        all_pairs = (
            (file, other_file)
            for file in image_files
            for other_file in image_files[image_files.index(file) + 1:]
        )
        logging.info('Filtering duplicates')
        matches = self.filter_matches(all_pairs)
        self.progress_bars.close()
        return matches

    def get_hashes(self, image_files: List[Path]) -> Dict[Path, ImageHash]:
        return {
            file: image_hash for file, image_hash in self.precalculate_hashes(image_files)
            if image_hash is not None
        }

    def filter_matches(self, all_pairs: Iterator[ImagePair]) -> Results:
        self.progress_bars.create_filter_bar(len(self.precalculated_hashes))
        return [
            (file, other_file) for file, other_file in all_pairs
            if self.are_images_equal(file, other_file)
        ]

    def are_images_equal(self, file: Path, other_file: Path) -> bool:
        self.progress_bars.update_filter()
        hash_distance = self.precalculated_hashes[file] - self.precalculated_hashes[other_file]
        logging.debug(
            '%-30s - %-30s = %d', file.stem, other_file.stem, hash_distance
        )
        return hash_distance <= self.max_distance


class DictImagePairFinder(ImagePairFinder):
    """
    Works this way only if max_distance == 0, otherwise needs to do an O(N^2) comparison between all
    image hashes
    """
    def __init__(  # pylint: disable = too-many-arguments
            self, files: List[Path], hash_algorithm: HashFunction,
            options: PairFinderOptions = PairFinderOptions(),
            hash_store: Optional[Cache] = None
    ) -> None:
        super().__init__(files, hash_algorithm, options, hash_store)
        if options.max_distance != 0:
            raise ValueError('DictImagePairFinder only works if max_distance == 0!')
        self.precalculated_hashes = self.get_hashes(files)
        self.progress_bars.close_reader()

    def get_pairs(self) -> Results:
        results = [
            pair
            for result in self.precalculated_hashes.values()
            for pair in combinations(list(result), 2)
            if len(result) > 1
        ]
        return results

    def get_hashes(self, image_files: List[Path]) -> Dict[ImageHash, Tuple[Path, ...]]:
        hash_dict: Dict[ImageHash, Tuple[Path, ...]] = defaultdict(tuple)
        for file in image_files:
            image_hash = self.get_hash(file)
            if len(image_hash) > 1 and image_hash[1] is not None:
                hash_dict[image_hash[1]] = hash_dict[image_hash[1]] + (file,)
        return hash_dict


class ParallelImagePairFinder(SerialImagePairFinder):
    def precalculate_hashes(self, image_files: List[Path]) -> List[CacheEntry]:
        with Pool() as pool:
            return pool.map(self.get_hash, image_files)


class ParallelFilteringImagePairFinder(ParallelImagePairFinder):
    """
    Not using this class at the moment since it seems too much trouble to get it to work correctly.
    See https://stackoverflow.com/a/44186168 for the reasons why.
    Keeping the code around anyway in case I want to tackle the problem of running the filtering in
    parallel one day. In that case ImagePairFinder.create() needs to return a
    ParallelFilteringImagePairFinder to utilize it.
    """
    CHUNK_SIZE = 100

    def filter_matches(self, all_pairs: Iterator[ImagePair]) -> Results:
        pairs = list(all_pairs)
        with Pool() as pool:
            to_keep = pool.starmap(
                self.are_images_equal, pairs, chunksize=self.CHUNK_SIZE
            )
        return [c for c, keep in zip(pairs, to_keep) if keep]



__author__ = 'Lene Preuss <lene.preuss@gmail.com>'

import logging
import os
from argparse import Namespace
from functools import lru_cache
from hashlib import sha256
from pathlib import Path
from shlex import quote
from subprocess import call  # noqa: S404
from typing import Any, Callable, Dict, List, Optional

import imagehash

from duplicate_images.common import path_with_parent
from duplicate_images.function_types import ActionFunction, HashFunction, ImagePair

__all__ = ['call', 'get_hash_size_kwargs', 'IMAGE_HASH_ALGORITHM', 'ACTIONS_ON_EQUALITY']


@lru_cache(maxsize=None)
def get_size(file: Path) -> int:
    return file.stat().st_size


@lru_cache(maxsize=None)
def get_hash(file: Path) -> str:
    return sha256(file.open('rb').read()).hexdigest()


def compare_exactly(file: Path, other_file: Path) -> bool:
    """Returns True if file and other_file are exactly exactly_equal"""
    return get_size(other_file) == get_size(file) and get_hash(file) == get_hash(other_file)


def ascending_by_size(pair: ImagePair) -> List[Path]:
    return sorted(pair, key=lambda path: path.stat().st_size)


def delete_with_log_message(file: Path) -> None:
    file.unlink()
    logging.info('Deleted %s', path_with_parent(file))


def shell_exec(args: Namespace, pair: ImagePair) -> None:
    num = 0
    cmd = args.exec
    for path in pair:
        num = num + 1
        cmd = cmd.replace(f"{'{'}{num}{'}'}", f'"{path}"')

    os.system(cmd)


def get_hash_size_kwargs(algorithm: HashFunction, size: Optional[int]) -> Dict[str, int]:
    if size is None:
        return ALGORITHM_DEFAULTS.get(algorithm, {'hash_size': 8})
    kwarg = next(iter(ALGORITHM_DEFAULTS.get(algorithm, {'hash_size': 8})))
    return {kwarg: size}


IMAGE_HASH_ALGORITHM = {
    'ahash': imagehash.average_hash,
    'phash': imagehash.phash,
    'phash_simple': imagehash.phash_simple,
    'dhash': imagehash.dhash,
    'dhash_vertical': imagehash.dhash_vertical,
    'whash': imagehash.whash,
    'colorhash': imagehash.colorhash,
}  # type: Dict[str, Callable[[Any], imagehash.ImageHash]]

ALGORITHM_DEFAULTS = {
    imagehash.average_hash: {'hash_size': 8},
    imagehash.phash: {'hash_size': 8},
    imagehash.phash_simple: {'hash_size': 8},
    imagehash.dhash: {'hash_size': 8},
    imagehash.dhash_vertical: {'hash_size': 8},
    imagehash.whash: {'hash_size': 8},
    imagehash.colorhash: {'binbits': 3},
}

ACTIONS_ON_EQUALITY: Dict[str, ActionFunction] = {
    'delete-first': lambda args, pair: delete_with_log_message(pair[0]),
    'd1': lambda args, pair: delete_with_log_message(pair[0]),
    'delete-second': lambda args, pair: delete_with_log_message(pair[1]),
    'd2': lambda args, pair: delete_with_log_message(pair[1]),
    'delete-bigger': lambda args, pair: delete_with_log_message(ascending_by_size(pair)[-1]),
    'd>': lambda args, pair: delete_with_log_message(ascending_by_size(pair)[-1]),
    'delete-smaller': lambda args, pair: delete_with_log_message(ascending_by_size(pair)[0]),
    'd<': lambda args, pair: delete_with_log_message(ascending_by_size(pair)[0]),
    'eog': lambda args, pair: call(['eog'] + [str(pic) for pic in pair]),  # noqa: S603
    'xv': lambda args, pair: call(['xv', '-nolim'] + [str(pic) for pic in pair]),  # noqa: S603
    'print': lambda args, pair: print(pair[0], pair[1]),
    'print_inline': lambda args, pair: print(pair[0], pair[1], end=' '),
    'quote': lambda args, pair: print(f'{quote(str(pair[0]))} {quote(str(pair[1]))}'),
    'quote_inline': lambda args, pair: print(
        f'{quote(str(pair[0]))} {quote(str(pair[1]))}', end=' '
    ),
    'exec': lambda args, pair: shell_exec(args, pair),  # pylint: disable=unnecessary-lambda
    'none': lambda args, pair: None
}



__author__ = 'Lene Preuss <lene.preuss@gmail.com>'

import logging
from argparse import Namespace

import coloredlogs


def setup_logging(args: Namespace) -> None:
    log_level = logging.DEBUG if args.debug else logging.INFO
    for _ in range(args.quiet):
        log_level += (logging.INFO - logging.DEBUG)
    coloredlogs.install(
        level=log_level, fmt='%(asctime)s %(levelname)s: %(message)s',
        datefmt='%H:%M:%S'
    )



__author__ = 'Lene Preuss <lene.preuss@gmail.com>'

from typing import Optional, Union

from tqdm import tqdm


class NullProgressBarManager:

    def __init__(self) -> None:
        pass

    def create_filter_bar(self, hashes_length: int) -> None:
        pass

    def update_reader(self) -> None:
        pass

    def update_filter(self) -> None:
        pass

    def close_reader(self) -> None:
        pass

    def close(self) -> None:
        pass


class ProgressBarManager:

    @classmethod
    def create(
            cls, files_length: int, active: bool = False
    ) -> Union['ProgressBarManager', NullProgressBarManager]:
        return ProgressBarManager(files_length) if active else NullProgressBarManager()

    def __init__(self, files_length: int) -> None:
        self.reader_progress: Optional[tqdm] = tqdm(
            total=files_length, miniters=max(files_length / 100, 5), smoothing=0.1, unit=''
        )
        self.filter_progress: Optional[tqdm] = None

    def create_filter_bar(self, hashes_length: int) -> None:
        self.close_reader()
        total_items = int(hashes_length * (hashes_length - 1) / 2)
        self.filter_progress = tqdm(
            total=total_items, unit='',
            unit_scale=True, miniters=max(total_items / 5000, 20000)
        )

    def update_reader(self) -> None:
        if self.reader_progress is not None:
            self.reader_progress.update(1)

    def update_filter(self) -> None:
        if self.filter_progress is not None:
            self.filter_progress.update(1)

    def close_reader(self) -> None:
        if self.reader_progress is not None:
            self.reader_progress.close()

    def close(self) -> None:
        if self.filter_progress is not None:
            self.filter_progress.close()
